/*
 *	KhepComMac.c
 *
 *	Khepera remote control via serial link on the Macintosh
 *
 *	(c) 1995-1999, Yves Piguet <yves.piguet@calerga.com>
 */

//#define EXTERN_API(_type) extern pascal _type

#include "KhepCom.h"
#include <Events.h>
#include <Devices.h>
#include <Serial.h>
#include <CRMSerialDevices.h>
#include <CommResources.h>

/*
 *	KPortRef *port:
 *	port->priv[0]: input device descriptor returned by OpenDriver
 *	port->priv[1]: output device descriptor
 *	port->priv[2]: timeout in ticks (1/60 s), or -1 if none
 *	port->priv[3]: unused
 *	 
*/

static int GetSerialDevice(long i, Str255 name, Str255 in, Str255 out)
	// retrieve the name of serial device i (i >= 0)
	// ret. 1 if ok, 0 if none
{
	// implementation: see
	// http://developer.apple.com/technotes/tn/tn1119.html
	CRMRec commRecord;
	CRMRecPtr thisCommRecord;
	CRMSerialPtr serialPtr;
	
	if (i < 0)
		return 0;
	
	(void)InitCRM();
	
	// First set up commRecord to specify that
	// we're interested in serial devices.
	
	commRecord.crmDeviceType = crmSerialDevice;
	commRecord.crmDeviceID   = 0;
  
	// Now repeatedly call CRMSearch to iterate
	// through all the serial ports.
  
	for (thisCommRecord = &commRecord; i >= 0 && thisCommRecord; i--)
		thisCommRecord = (CRMRecPtr)CRMSearch((CRMRecPtr)thisCommRecord);
	
	if (thisCommRecord)
	{
		// Once we a have a CRMRec for the serial port,
		// we must cast the crmAttributes field to
		// a CRMSerialPtr to access to serial-specific
		// fields about the port.
  
		serialPtr = (CRMSerialPtr) thisCommRecord->crmAttributes;
  
		// get the information about the port.
		if (name)
			for (i = 0; i <= (*serialPtr->name)[0]; i++)
				name[i] = (*serialPtr->name)[i];
		if (in)
			for (i = 0; i <= (*serialPtr->inputDriverName)[0]; i++)
				in[i] = (*serialPtr->inputDriverName)[i];
		if (out)
			for (i = 0; i <= (*serialPtr->outputDriverName)[0]; i++)
				out[i] = (*serialPtr->outputDriverName)[i];
		return 1;
	}
	else
		return 0;
}

int KGetSerialDeviceName(long i, char *name, long *nameLength)
	// retrieve the name of serial device i (i >= 0)
	// ret. 1 if ok, 0 if none
{
	Str255 nameStr;
	
	if (!GetSerialDevice(i, nameStr, 0, 0))
		return 0;
	for (i = 0; i < nameStr[0] && i < kNameMaxLength - 1; i++)
		name[i] = (char)nameStr[i + 1];
	name[i] = '\0';
	if (nameLength)
		*nameLength = i;
	
	return 1;
}

int KOpenSerial(long port, long brate, long timeout, KPortRef *portRef)
	// open and configure the serial port for communicating with Khepera
	// same parameters as SendCommandToKhepera; return a port reference in *portRef
	//  and an error code

{
	Str63 inName, outName;
	short in, out;
	SerShk flags;
	short serConfig;
	short higherSpeed = 0;	// non-zero -> special configuration,
							// not directly supported by SerReset()
							// (cf. Apple Technote 1018)
	OSErr err;
	long n, count;
	long tks;
	char dummy;
	
	timeout /= 1000 / 60;	// from ms to ticks (1/60 s)
		
	flags.fXOn = 0;
	flags.fCTS = 0;
	flags.xOn = 0;
	flags.xOff = 0;
	flags.errs = 0;
	flags.evts = 0;
	flags.fInX = 0;
	flags.fDTR = 0;
	
	if (!GetSerialDevice(port, 0, inName, outName))
		return kKhepComBadPortErr;
	
	switch (brate)
	{
		case 150:
			serConfig = baud150 | stop20 | noParity | data8;
			break;
		case 300:
			serConfig = baud300 | stop20 | noParity | data8;
			break;
		case 600:
			serConfig = baud600 | stop20 | noParity | data8;
			break;
		case 1200:
			serConfig = baud1200 | stop20 | noParity | data8;
			break;
		case 1800:
			serConfig = baud1800 | stop20 | noParity | data8;
			break;
		case 2400:
			serConfig = baud2400 | stop20 | noParity | data8;
			break;
		case 4800:
			serConfig = baud4800 | stop20 | noParity | data8;
			break;
		case 7200:
			serConfig = baud7200 | stop20 | noParity | data8;
			break;
		case 9600:
			serConfig = baud9600 | stop20 | noParity | data8;
			break;
		case 14400:
			serConfig = baud14400 | stop20 | noParity | data8;
			break;
		case 19200:
			serConfig = baud19200 | stop20 | noParity | data8;
			break;
		case 28800:
			serConfig = baud28800 | stop20 | noParity | data8;
			break;
		case 38400:
			serConfig = baud38400 | stop20 | noParity | data8;
			break;
		case 57600:
			serConfig = baud57600 | stop20 | noParity | data8;
			break;
		case 115200:
			serConfig = baud57600 | stop20 | noParity | data8;	// some dummy baud rate
			higherSpeed = 115;									// csCode for PBStatus
			break;
		case 230400:
			serConfig = baud57600 | stop20 | noParity | data8;
			higherSpeed = 230;
			break;	
		default:
			return kKhepComBadBaudRateErr;
	}
	
	err = OpenDriver(inName, &in);
	if (err != noErr)
		return kKhepDeviceErr;
	err = OpenDriver(outName, &out);
	if (err != noErr)
	{
		CloseDriver(in);
		return kKhepDeviceErr;
	}
	SerReset(out, serConfig);
	SerReset(in, serConfig);
	SerHShake(out, &flags);
	SerHShake(in, &flags);
	
	if (higherSpeed != 0)
	{
		CntrlParam param;
		
		param.ioCompletion = 0;
		param.ioVRefNum = 0;
		param.ioCRefNum = out;
		param.csCode = higherSpeed;
		if (PBControl((ParmBlkPtr)&param, 0) != noErr
					|| (param.ioCRefNum = in, PBControl((ParmBlkPtr)&param, 0) != noErr))
		{
			CloseDriver(in);
			CloseDriver(out);
			return kKhepDeviceErr;
		}
	}
	
	portRef->priv[0] = in;
	portRef->priv[1] = out;
	portRef->priv[2] = timeout;
	portRef->priv[3] = 0;
	
	// flush input
	
	for (tks = TickCount(); timeout < 0 || TickCount() < tks + timeout;)
	{
		SerGetBuf(in, &n);
		count = 1;
		if (n > 0)
			err = FSRead(in, &count, &dummy);
		if (n <= 0 || err != noErr)
			break;
	}
	
	return err;
}

void KCloseSerial(KPortRef *portRef)
	// close the serial port
{
	short in, out;
	
	in = portRef->priv[0];
	out = portRef->priv[1];
	
	KillIO(in);
	CloseDriver(in);
	KillIO(out);
	CloseDriver(out);
}

void KFlush(KPortRef *portRef)
{
	short in;
	long tks, timeout, n, count;
	char dummy;
	OSErr err;
	
	in = portRef->priv[0];
	timeout = portRef->priv[2];
	
	for (tks = TickCount(); timeout < 0 || TickCount() < tks + timeout;)
	{
		SerGetBuf(in, &n);
		count = 1;
		if (n > 0)
			err = FSRead(in, &count, &dummy);
		if (n <= 0 || err != noErr)
			break;
	}
}

int KSend(KPortRef *portRef, const char *buf, long length)
{
	if (length > 0)
	{
		short out;
		
		out = portRef->priv[1];
		if (FSWrite(out, &length, buf) != noErr)
			return kKhepComErr;
	}
	
	return kKhepNoErr;
}

int KReceive(KPortRef *portRef, char *buf, long maxLength, long *length)
{
	short in;
	long timeout;
	OSErr err;
	long n, count, i;
	long tks;
	
	in = portRef->priv[0];
	timeout = portRef->priv[2];
	
	tks = TickCount();
	
	for (i = 0; i < maxLength && (timeout < 0 || TickCount() < tks + timeout); )
	{
		SerGetBuf(in, &n);
		if (n > 0)
		{
//DebugStr("\pNon-empty ser buffer");
			count = n + i > maxLength ? maxLength - i : n;
			err = FSRead(in, &count, buf + i);
			if (err != noErr)
				goto end;
			i += count;
		}
	}
	
	*length = i;
	
end:
	if (*length > 0 || timeout < 0 || TickCount() < tks + timeout)
		return err == noErr ? kKhepNoErr : kKhepComErr;
	else
		return kKhepComTimeoutErr;
}

int KSendCommand(KPortRef *portRef,	// as returned by KOpenSerial
						const char *command, long cmdLength,
						char **answer, long answerMaxLength, long *answerLength,
						int stopAtCr,	// if false, read until timeout
						reallocPtr myRealloc)
	// use myRealloc if answerMaxLength is not large enough
	// answer is null-terminated (C-string)
{
	short in, out;
	long timeout;
	OSErr err;
	long n, count, i;
	long tks;
	char *oldPtr;
	
	in = portRef->priv[0];
	out = portRef->priv[1];
	timeout = portRef->priv[2];
	
	tks = TickCount();
	
	if (cmdLength > 0)
	{
		// send command
		
		count = cmdLength;
		err = FSWrite(out, &count, command);
		if (err != noErr)
			goto end;
	}
	
	// read reply
	
	for (i = 0; timeout < 0 || TickCount() < tks + timeout; )
	{
		SerGetBuf(in, &n);
		if (n > 0)
		{
			if (n + i > answerMaxLength && myRealloc)
			{
				oldPtr = *answer;
				*answer = myRealloc(oldPtr,
								answerMaxLength < 256 ? 512 : 2 * answerMaxLength);
				
				if (*answer)	// ok
				{
					answerMaxLength = answerMaxLength < 256 ? 512 : 2 * answerMaxLength;
					count = n;
				}
				else
				{
					*answer = oldPtr;
					count = answerMaxLength - i;
				}
			}
			else
				count = n + i > answerMaxLength ? answerMaxLength - i : n;
			err = FSRead(in, &count, *answer + i);
			if (err != noErr)
				goto end;
			i += count;
			if (stopAtCr && (*answer)[i - 1] == '\n')
			{
				(*answer)[i] = '\0';
				*answerLength = i - 1;
				goto end;
			}
		}
	}
	
	(*answer)[i - 1] = '\0';
	*answerLength = i - 1;
end:
	
	// remove cr's
	for (n = i = 0; n < *answerLength; n++)
		if ((*answer)[n] != '\r')
			(*answer)[i++] = (*answer)[n];
	*answerLength = i;
	(*answer)[i] = '\0';
	
	if (!stopAtCr || timeout < 0 || TickCount() < tks + timeout)
		return err == noErr ? kKhepNoErr : kKhepComErr;
	else
		return kKhepComTimeoutErr;
}

int KOpenSendClose(long port, long brate, long timeout,
						char *command, long cmdLength,
						char **answer, long answerMaxLength, long *answerLength,
						int stopAtCr,	// if false, read until timeout
						reallocPtr myRealloc)
	// perform KOpenSerial, KSendCommand, and KCloseSerial
{
	KPortRef p;
	int err;
	
	err = KOpenSerial(port, brate, timeout, &p);
	if (err != kKhepNoErr)
		return kKhepNoErr;
	err = KSendCommand(&p, command, cmdLength,
						answer, answerMaxLength, answerLength,
						stopAtCr, myRealloc);
	KCloseSerial(&p);
	return err;
}
