/*
	ksend		Send a command to Khepera
	
	The calling syntax is:
	
		result = ksend(cmd)
	
	or
	
		result = khepcmd(cmd, ref, stopAtCr)
	
	Arguments:
	
	cmd:		the command to send (without the carriage return) as a string
	ref:		the reference to the serial port obtained with kopen
	stopAtCr:	true if the end of the reply is indicated by a cr or lf character,
				false multiple lines are expected and the timeout is used
	result:		the reply of Khepera
	

	Under Windows, compile by typing in Matlab's command line

		mex -output ksend KSend.c KhepComWin.c
	
	Copyright 1998, Yves Piguet
	All Righs Reserved
*/

//#include "CodeWarrior.MEX.prefix"

#include "mex.h"

#include "KhepCom.h"



#define dynamic_reply	// COMMENT OUT THIS LINE TO HAVE A FIXED-SIZE REPLY BUFFER

/* Input Arguments */

#define	CMD_IN		prhs[0]
#define	PORTREF_IN	prhs[1]
#define	STOPATCR_IN	prhs[2]

/* Output Arguments */

#define	RESULT_OUT	plhs[0]

// malloc or realloc, sent to SendCommandToKhepera()

static void *myRealloc(void *ptr, long size)
{
	if (ptr)
		return mxRealloc(ptr, (size_t)size);
	else
		return mxMalloc((size_t)size);
}

// Main Function

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
#ifdef dynamic_reply
	char *string;
#else
	char string[512];
#endif
	char *answer;
	long cmdLength, answerLength;
	KPortRef portRef;
	double *ref;	// same, in the input argument
	int stopAtCr;
	int err;
	
	/* Checks for proper number of arguments */
	
	if (nrhs < 2 || nrhs > 3 || nlhs > 1)
		mexErrMsgTxt("KSEND syntax: reply = ksend(cmd,ref,stopAtCr)");
	
	/* Gets the parameters */
	
	if (!mxIsChar(CMD_IN) || mxIsSparse(CMD_IN) || mxGetM(CMD_IN) != 1)
		mexErrMsgTxt("KSEND first arg. must be a string");
	
	cmdLength = mxGetN(CMD_IN);
#ifdef dynamic_reply
	string = mxMalloc((size_t)(cmdLength + 1));
	if (!string)
		mexErrMsgTxt("KSEND not enough memory");
#else
	if (cmdLength > 510)
		mexErrMsgTxt("KSEND the command is longer than 510 characters");
#endif
	mxGetString(CMD_IN, string, cmdLength + 1);
	
	if(!mxIsDouble(PORTREF_IN) || mxIsComplex(PORTREF_IN) ||
								mxGetM(PORTREF_IN) != 1 || mxGetN(PORTREF_IN) != 4) 
		mexErrMsgTxt("KSEND arg. must be a row vector of four numbers");
	ref = mxGetPr(PORTREF_IN);
	portRef.priv[0] = ref[0];
	portRef.priv[1] = ref[1];
	portRef.priv[2] = ref[2];
	portRef.priv[3] = ref[3];
	
	if (nrhs >= 3)
	{
		if(!mxIsDouble(STOPATCR_IN) || mxIsComplex(STOPATCR_IN) ||
									mxGetM(STOPATCR_IN) != 1 || mxGetN(STOPATCR_IN) != 1) 
			mexErrMsgTxt("KSEND second arg. must be a scalar boolean value");
		stopAtCr = mxGetPr(STOPATCR_IN)[0] != 0;
	}
	else
		stopAtCr = 1;
	
#ifdef dynamic_reply
	answer = 0;
	err = KSendCommand(&portRef,
					string, cmdLength,
					&answer, 0, &answerLength, stopAtCr,
					myRealloc);
	
	if (err == kKhepNoErr && nlhs >= 1)
		RESULT_OUT = mxCreateString(answer ? answer : "");
	
	if (answer)
		mxFree(answer);
#else
	{
		char reply[2048];
		answer = reply;
		err = KSendCommand(&portRef,
						string, cmdLength,
						&answer, 1024, &answerLength, stopAtCr,
						0);
		
		if (err == kKhepNoErr && nlhs >= 1)
			RESULT_OUT = mxCreateString(answer ? answer : "");
	}
#endif
	
#ifdef dynamic_reply
	mxFree(string);
#endif
	
	switch (err)
	{
		case kKhepDeviceErr:
			mexErrMsgTxt("KSEND device error");
			break;
		case kKhepComErr:
			mexErrMsgTxt("KSEND read/write error");
			break;
		case kKhepComTimeoutErr:
			mexErrMsgTxt("KSEND timeout error");
			break;
		case kKhepComBadPortErr:
			mexErrMsgTxt("KSEND unsupported serial port");
			break;
		case kKhepComBadBaudRateErr:
			mexErrMsgTxt("KSEND unsupported baud rate");
			break;
	}
}
