/*
 *	SQKhep.c
 *
 *	Functions to communicate with Khepera
 *
 *	Written by Yves Piguet
 *	Copyright 1999, Yves Piguet and K-Team S.A.
 */

#include "SQKhep.h"
#include "KhepCom.h"

static int isOpen = 0;
static KPortRef thePortRef;

LMEErr XF_KSerialName(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)
{
	LMEFloat port;
	LMEText name;
	LMEInt nameLength;
	LMEErr err;
	(void)nargin, nargout, user;
	
	err = LME_PopReal(lme, &port);
	if (err != kLMEErrOk)
		return err;
	
	nameLength = -1;
	err = LME_BeginPushString(lme, &nameLength, &name);
	if (err != kLMEErrOk)
		return err;
	
	if (KGetSerialDeviceName((long)port, name, &nameLength))	// ok
	{
		LME_EndPushString(lme, nameLength);
		return kLMEErrOk;
	}
	else
	{
		LME_RemoveObjects(lme, 0, 1);
		return LME_PushString(lme, 0, "");
	}
}

LMEErr XF_KOpen(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)
{
	LMEInt nl, nc;
	LMEFloat *re;
	LMEErr err;
	int khepErr;
	KPortRef portRef;
	(void)nargin, nargout, user;
	
	LME_PopMatrix(lme, &nl, &nc, &re, 0, 0, 0);
	if (nl * nc != 3)
		return kLMEErrBadSize;
	
	khepErr = KOpenSerial((int)re[0], (int)re[1],
					re[2] < 0 ? 60000L : re[2] > 60 ? 60000L : (long)(re[2] * 1000),
					&portRef);
	
	switch (khepErr)
	{
		case kKhepNoErr:
			err = LME_PushMatrix(lme, 1, 4, 0, &re, 0, 0, 0);
			if (err != kLMEErrOk)
				return err;
			re[0] = portRef.priv[0];
			re[1] = portRef.priv[1];
			re[2] = portRef.priv[2];
			re[3] = portRef.priv[3];
			break;
		case kKhepDeviceErr:
			return kLMEErrBadAccess;
		case kKhepComErr:
			return kLMEErrBadAccess;
		case kKhepComTimeoutErr:
			return kLMEErrTimeout;
		case kKhepComBadPortErr:
			return kLMEErrFileNotFound;
		case kKhepComBadBaudRateErr:
			return kLMEErrBadArg;
	}
	
	isOpen = 1;
	thePortRef = portRef;
	return kLMEErrOk;
}

LMEErr XF_KSend(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)
{
	LMEInt nl, nc, cmdLength, replyLength;
	LMEFloat x, *re;
	LMEBoolean stopAtCr;
	LMEText cmd, reply;
	LMEErr err;
	int khepErr;
	KPortRef portRef;
	(void)nargout, user;
	
	// stopAtCr (default = true)
	if (nargin >= 3)
	{
		err = LME_PopReal(lme, &x);
		if (err != kLMEErrOk)
			return err;
		stopAtCr = (LMEBoolean)x;
	}
	else
		stopAtCr = 1;
	
	// ref
	LME_PopMatrix(lme, &nl, &nc, &re, 0, 0, 0);
	if (nl * nc != 4)
		return kLMEErrBadSize;
	portRef.priv[0] = re[0];
	portRef.priv[1] = re[1];
	portRef.priv[2] = re[2];
	portRef.priv[3] = re[3];
	
	// cmd
	err = LME_ConvertStringInPlace(lme, 0, &cmd, &cmdLength);
	if (err != kLMEErrOk)
		return err;
	
	// send command
	replyLength = -1;
	err = LME_BeginPushString(lme, &replyLength, &reply);
	if (err != kLMEErrOk)
		return err;
	khepErr = KSendCommand(&portRef, cmd, cmdLength,
						&reply, replyLength, &replyLength, stopAtCr,
						0);
	if (khepErr == kKhepNoErr)
	{
		LME_EndPushString(lme, replyLength);
		LME_RemoveObjects(lme, 1, 1);
		return kLMEErrOk;
	}
	else
	{
		LME_RemoveObjects(lme, 0, 2);
		switch (khepErr)
		{
			case kKhepDeviceErr:
				return kLMEErrBadAccess;
			case kKhepComErr:
				return kLMEErrBadAccess;
			case kKhepComTimeoutErr:
				return kLMEErrTimeout;
			case kKhepComBadPortErr:
				return kLMEErrFileNotFound;
			case kKhepComBadBaudRateErr:
				return kLMEErrBadArg;
			default:
				return kLMEErrBadAccess;
		}
	}
}

LMEErr XF_KFlush(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)
{
	LMEInt nl, nc;
	LMEFloat *re;
	KPortRef portRef;
	(void)nargin, nargout, user;
	
	// ref
	LME_PopMatrix(lme, &nl, &nc, &re, 0, 0, 0);
	if (nl * nc != 4)
		return kLMEErrBadSize;
	portRef.priv[0] = re[0];
	portRef.priv[1] = re[1];
	portRef.priv[2] = re[2];
	portRef.priv[3] = re[3];
	
	// flush
	KFlush(&portRef);
	
	return kLMEErrOk;
}

LMEErr XF_KClose(LMEPtr lme, LMEInt nargin, LMEInt nargout, void *user)
{
	LMEInt nl, nc;
	LMEFloat *re;
	KPortRef portRef;
	(void)nargin, nargout, user;
	
	LME_PopMatrix(lme, &nl, &nc, &re, 0, 0, 0);
	if (nl * nc != 4)
		return kLMEErrBadSize;
	
	portRef.priv[0] = re[0];
	portRef.priv[1] = re[1];
	portRef.priv[2] = re[2];
	portRef.priv[3] = re[3];
	
	KCloseSerial(&portRef);
	
	isOpen = 0;
	
	return kLMEErrOk;
}

void KShutdown(void)
	// close the connection if it's open
{
	if (isOpen)
	{
		KCloseSerial(&thePortRef);
		isOpen = 0;
	}
}
